def encode_tool_call_id(tool_call_id: str) -> str:
    """Encode the tool call id so that it fits within a TCSS id.

    Args:
        tool_call_id: Raw tool call id.

    Returns:
        Tool call usable as widget id.
    """
    hex_tool_call_id = "".join(f"{ord(character):2X}" for character in tool_call_id)
    encoded_tool_call_id = f"tool-call-{hex_tool_call_id}"
    return encoded_tool_call_id



from dataclasses import dataclass
import mimetypes
from pathlib import Path


@dataclass
class Resource:
    root: Path
    path: Path
    mime_type: str
    text: str | None
    data: bytes | None


class ResourceError(Exception):
    """An error occurred reading a resource."""


class ResourceNotRelative(ResourceError):
    """Attempted to read a resource, not in the project directory."""


class ResourceReadError(ResourceError):
    """Failed to read the resource."""


def load_resource(root: Path, path: Path) -> Resource:
    """Load a resource from the project directory.

    Args:
        root: The project root.
        path: Relative path within project.

    Returns:
        A resource.
    """
    resource_path = root / path

    if not resource_path.is_relative_to(root):
        raise ResourceNotRelative("Resource path is not relative to project root.")

    mime_type, encoding = mimetypes.guess_file_type(resource_path)
    if mime_type is None:
        mime_type = "application/octet-stream"

    data: bytes | None
    text: str | None

    try:
        if encoding is not None:
            data = resource_path.read_bytes()
            text = None
        else:
            data = None
            text = resource_path.read_text(encoding, errors="replace")
    except FileNotFoundError:
        raise ResourceReadError(f"File not found {str(path)!r}")
    except Exception as error:
        raise ResourceReadError(f"Failed to read {str(path)!r}; {error}")

    resource = Resource(
        root,
        resource_path,
        mime_type=mime_type,
        text=text,
        data=data,
    )
    return resource



from time import monotonic

from textual.widget import Widget

from textual.content import Content
from textual.reactive import reactive


class StrikeText(Widget):
    DEFAULT_CSS = """
    StrikeText {
        height: auto;
    }
    """

    strike_time: reactive[float | None] = reactive(None)

    def __init__(
        self,
        content: Content,
        name: str | None = None,
        id: str | None = None,
        classes: str | None = None,
    ):
        self.content = content
        super().__init__(name=name, id=id, classes=classes)

    def strike(self) -> None:
        self.strike_time = monotonic()
        self.auto_refresh = 1 / 30

    def render(self) -> Content:
        content = self.content
        if self.strike_time is not None:
            position = int((monotonic() - self.strike_time) * 70)
            content = content.stylize("strike", 0, position)
            if position > len(content):
                self.auto_refresh = None
        return content


if __name__ == "__main__":
    from textual.app import App, ComposeResult
    from textual.widgets import Static

    class StrikeApp(App):
        CSS = """
        Screen {
            overflow: auto;
        }

        """
        BINDINGS = [("space", "strike", "Strike")]

        def compose(self) -> ComposeResult:
            for n in range(20):
                yield Static("HELLO")
            yield StrikeText(Content("Where there is a Will, there is a way"))
            for n in range(200):
                yield Static("World")

        def action_strike(self):
            self.query_one(StrikeText).strike()

    app = StrikeApp()
    app.run()



import re2 as re
from typing import Iterable


RE_MATCH_FILE_PROMPT = re.compile(r"@(\S+)|@\"(.*)\"")


def extract_paths_from_prompt(prompt: str) -> Iterable[tuple[str, int, int]]:
    """Find file syntax in prompts.

    Args:
        prompt: A line of prompt.

    Yields:
        A tuple of (PATH, START, END).
    """
    for match in RE_MATCH_FILE_PROMPT.finditer(prompt):
        path, quoted_path = match.groups()
        yield (path or quoted_path, match.start(0), match.end(0))



from textual.content import Content


def pill(text: Content | str, background: str, foreground: str) -> Content:
    """Format text as a pill (half block ends).

    Args:
        text: Pill contents as Content object or text.
        background: Background color.
        foreground: Foreground color.

    Returns:
        Pill content.
    """
    content = Content(text) if isinstance(text, str) else text
    main_style = f"{foreground} on {background}"
    end_style = f"{background} on transparent r"
    pill_content = Content.assemble(
        ("▌", end_style),
        content.stylize(main_style),
        ("▐", end_style),
    )
    return pill_content



from dataclasses import dataclass
from typing import Iterable

from textual.content import Content
from textual.dom import DOMNode
from textual.message import Message

from toad.menus import MenuItem
from toad.widgets.terminal import Terminal


class ShellTerminal(Terminal):
    """Subclass of Terminal used in the Shell view."""

    @dataclass
    class Terminate(Message):
        teminal: ShellTerminal

        @property
        def control(self) -> DOMNode:
            return self.terminal

    @dataclass
    class Interrupt(Message):
        teminal: ShellTerminal

        @property
        def control(self) -> DOMNode:
            return self.terminal

    def get_block_menu(self) -> Iterable[MenuItem]:
        if not self.is_finalized:
            yield MenuItem("Interrupt", "interrupt", "i")
            yield MenuItem("Focus", f"focus_block({self.id!r})", "f")

    def get_block_content(self, destination: str) -> str | None:
        return "\n".join(line.content.plain for line in self.state.buffer.lines)

    def on_mount(self) -> None:
        self.border_title = Content(self.name)



from textual.content import Content
from textual.css.styles import RulesMap
from textual.style import Style
from textual.visual import Visual, RenderOptions
from textual.strip import Strip

from itertools import zip_longest


class OptionContent(Visual):
    def __init__(self, option: str | Content, help: str | Content) -> None:
        self.option = Content(option) if isinstance(option, str) else option
        self.help = Content(help) if isinstance(help, str) else help
        self._label = Content(f"{option} {help}")

    def __str__(self) -> str:
        return str(self.option)

    def render_strips(
        self, width: int, height: int | None, style: Style, options: RenderOptions
    ) -> list[Strip]:
        option_strips = [
            Strip(
                self.option.render_segments(style), cell_length=self.option.cell_length
            )
        ]

        option_width = self.option.cell_length
        remaining_width = width - self.option.cell_length

        help_strips = self.help.render_strips(
            remaining_width, None, style, options=options
        )
        help_width = max(strip.cell_length for strip in help_strips)
        help_width = [strip.extend_cell_length(help_width) for strip in help_strips]

        strips: list[Strip] = []
        for option_strip, help_strip in zip_longest(option_strips, help_strips):
            if option_strip is None:
                option_strip = Strip.blank(option_width)
            assert isinstance(help_strip, Strip)
            strips.append(Strip.join([option_strip, help_strip]))
        return strips

    def get_optimal_width(self, rules: RulesMap, container_width: int) -> int:
        return self._label.get_optimal_width(rules, container_width)

    def get_height(self, rules: RulesMap, width: int) -> int:
        label_width = self.option.cell_length + 1
        height = self.help.get_height(rules, width - label_width)
        return height



from typing import Iterable
from textual.widgets import Static

from toad.menus import MenuItem


class Note(Static):
    DEFAULT_CLASSES = "block"

    def get_block_menu(self) -> Iterable[MenuItem]:
        return
        yield

    def get_block_content(self, destination: str) -> str | None:
        return str(self.render())

    def action_hello(self, message: str) -> None:
        self.notify(message, severity="warning")



from datetime import datetime, timezone
from typing import cast, TypedDict
from toad import paths

import aiosqlite


class Session(TypedDict):
    """Agent session fields."""

    id: int
    """Primary key."""
    agent: str
    """Title of the agent."""
    agent_identity: str
    """Agent identity."""
    agent_session_id: str
    """Agent's session id."""
    title: str
    """Title of session."""
    protocol: str
    """Protocol used."""
    promot_count: int
    """Number of prompts sent."""
    created_at: str
    """Time session was created."""
    last_used: str
    """Time sesison was last used."""
    meta_json: str
    """Text field containing JSON meta."""


class DB:
    """Toads database, for anything that isn't strictly configuration."""

    def __init__(self):
        self.path = paths.get_state() / "toad.db"

    def open(self) -> aiosqlite.Connection:
        return aiosqlite.connect(self.path)

    async def create(self) -> bool:
        """Create the tables if requried."""
        try:
            async with self.open() as db:
                await db.execute(
                    """
                    CREATE TABLE IF NOT EXISTS sessions (
                        id INTEGER PRIMARY KEY AUTOINCREMENT,
                        agent TEXT NOT NULL,
                        agent_identity TEXT NOT NULL,
                        agent_session_id TEXT NOT NULL,                                
                        title TEXT NOT NULL,      
                        protocol TEXT NOT NULL, 
                        prompt_count INTEGER DEFAULT 0,                
                        created_at TIMESTAMP DEFAULT CURRENT_TIMESTAMP,
                        last_used TIMESTAMP DEFAULT CURRENT_TIMESTAMP,
                        meta_json TEXT DEFAULT '{}'
                    )
                    """
                )
        except aiosqlite.Error:
            return False
        return True

    async def session_new(
        self,
        title: str,
        agent: str,
        agent_identity: str,
        agent_session_id: str,
        protocol: str = "acp",
    ) -> int | None:
        try:
            async with self.open() as db:
                cursor = await db.execute(
                    """
                    INSERT INTO sessions (title, agent, agent_identity, agent_session_id, protocol) VALUES (?, ?, ?, ?, ?)    
                    """,
                    (
                        title,
                        agent,
                        agent_identity,
                        agent_session_id,
                        protocol,
                    ),
                )
                await db.commit()
                return cursor.lastrowid
        except aiosqlite.Error:
            return None

    async def session_update_last_used(self, id: int) -> bool:
        """Update the last used timestamp.

        Args:
            id: Session ID.

        Returns:
            Boolenan that indicates success.
        """
        now_utc = datetime.now(timezone.utc)
        try:
            async with self.open() as db:
                await db.execute(
                    "UPDATE sessions SET last_used = ? WHERE id = ?",
                    (
                        now_utc.isoformat(),
                        id,
                    ),
                )
                await db.commit()
        except aiosqlite.Error:
            return False
        return True

    async def session_update_title(self, id: int, title: str) -> bool:
        """Update the last used timestamp.

        Args:
            id: Session ID.
            title: New title.

        Returns:
            Boolenan that indicates success.
        """
        try:
            async with self.open() as db:
                await db.execute(
                    "UPDATE sessions SET title = ? WHERE id = ?",
                    (
                        title,
                        id,
                    ),
                )
                await db.commit()
        except aiosqlite.Error:
            return False
        return True

    async def session_get(self, id: int) -> Session | None:
        """Get a sesison from its ID (PK).

        Args:
            session_id: The ID field (PK, not the agent_session_id)

        Returns:
            A Session if one is found, or `None`.
        """
        try:
            async with self.open() as db:
                db.row_factory = aiosqlite.Row
                cursor = await db.execute("SELECT * from sessions WHERE id = ?", (id,))
                row = await cursor.fetchone()
        except aiosqlite.Error:
            return None
        if row is None:
            return None
        session = cast(Session, dict(row))
        return session

    async def session_get_recent(self, max_results: int = 100) -> list[Session] | None:
        """Get the most recent sessions."""
        try:
            async with self.open() as db:
                db.row_factory = aiosqlite.Row
                cursor = await db.execute(
                    """SELECT * from sessions
                    ORDER BY last_used DESC
                    LIMIT ?""",
                    (max_results,),
                )
                rows = await cursor.fetchall()
        except aiosqlite.Error:
            return None
        sessions = [cast(Session, dict(row)) for row in rows]
        return sessions



from typing import NamedTuple

from rich.segment import Segment
from rich.color import Color as RichColor
from rich.style import Style as RichStyle

from textual import events
from textual.color import Color
from textual.content import Content
from textual.geometry import NULL_SIZE, Offset
from textual.reactive import reactive, var
from textual.strip import Strip
from textual.app import App, ComposeResult
from textual.widget import Widget
from textual.timer import Timer


COLORS = [
    Color.parse(color).rgb
    for color in [
        "#881177",
        "#aa3355",
        "#cc6666",
        "#ee9944",
        "#eedd00",
        "#99dd55",
        "#44dd88",
        "#22ccbb",
        "#00bbcc",
        "#0099cc",
        "#3366bb",
        "#663399",
    ]
]


class MandelbrotRegion(NamedTuple):
    """Defines the extents of the mandelbrot set."""

    x_min: float
    x_max: float
    y_min: float
    y_max: float

    def zoom(
        self, focal_x: float, focal_y: float, zoom_factor: float
    ) -> "MandelbrotRegion":
        """
        Return a new region zoomed in or out from a focal point.

        Args:
            focal_x: X coordinate of the point to zoom around (in complex plane coordinates)
            focal_y: Y coordinate of the point to zoom around (in complex plane coordinates)
            zoom_factor: Zoom factor (>1 to zoom in, <1 to zoom out, =1 for no change)

        Returns:
            A new MandelbrotRegion with the focal point at the same relative position
        """
        # Calculate current dimensions
        width = self.x_max - self.x_min
        height = self.y_max - self.y_min

        # Calculate new dimensions
        new_width = width / zoom_factor
        new_height = height / zoom_factor

        # Calculate focal point's relative position in current region
        fx = (focal_x - self.x_min) / width
        fy = (focal_y - self.y_min) / height

        # Calculate new bounds maintaining the focal point's relative position
        new_x_min = focal_x - fx * new_width
        new_x_max = focal_x + (1 - fx) * new_width
        new_y_min = focal_y - fy * new_height
        new_y_max = focal_y + (1 - fy) * new_height

        return MandelbrotRegion(new_x_min, new_x_max, new_y_min, new_y_max)


class Mandelbrot(Widget):
    ALLOW_SELECT = False
    DEFAULT_CSS = """
    Mandelbrot {        
        border: block black 20%;
        text-wrap: nowrap;
        text-overflow: clip;
        overflow: hidden;
    }
    """

    set_region = reactive(MandelbrotRegion(-2, 1.0, -1.0, 1.0), init=False)
    max_iterations = var(64)
    rendered_size = var(NULL_SIZE)
    rendered_set = var(Content(""))
    zoom_position = var(Offset(0, 0))
    zoom_timer: var[Timer | None] = var(None)
    zoom_scale = var(0.99)

    BRAILLE_CHARACTERS = [chr(0x2800 + i) for i in range(256)]

    # (BIT, X_OFFSET, Y_OFFSET)
    PATCH_COORDS = [
        (1, 0, 0),
        (2, 0, 1),
        (4, 0, 2),
        (8, 1, 0),
        (16, 1, 1),
        (32, 1, 2),
        (64, 0, 3),
        (128, 1, 3),
    ]

    def __init__(
        self,
        name: str | None = None,
        id: str | None = None,
        classes: str | None = None,
    ) -> None:
        self._strip_cache: dict[int, Strip] = {}
        super().__init__(name=name, id=id, classes=classes)

    @staticmethod
    def mandelbrot(c_real: float, c_imag: float, max_iterations: int):
        """
        Determine the smooth iteration count for a point in the Mandelbrot set.
        Uses continuous (smooth) iteration counting for better detail outside the set.

        Args:
            c_real: The real part of the complex number.
            c_imag: The imaginary part of the complex number.

        Returns:
            A float representing the smooth iteration count, or MAX_ITER for points in the set.
        """
        # Early escape: check if point is in main cardioid
        # The main cardioid can be detected with: q(q + (x - 1/4)) < 1/4 * y^2
        # where q = (x - 1/4)^2 + y^2
        x_shifted = c_real - 0.25
        q = x_shifted * x_shifted + c_imag * c_imag
        if q * (q + x_shifted) < 0.25 * c_imag * c_imag:
            return max_iterations

        # Early escape: check if point is in period-2 bulb
        # The period-2 bulb is the circle: (x + 1)^2 + y^2 < 1/16
        x_plus_one = c_real + 1.0
        if x_plus_one * x_plus_one + c_imag * c_imag < 0.0625:
            return max_iterations

        z_real = 0.0
        z_imag = 0.0
        for i in range(max_iterations):
            z_real_new = z_real * z_real - z_imag * z_imag + c_real
            z_imag_new = 2 * z_real * z_imag + c_imag
            z_real = z_real_new
            z_imag = z_imag_new
            if z_real * z_real + z_imag * z_imag > 4:
                return i
        return max_iterations

    def on_mount(self):
        self.call_after_refresh(self.refresh)

    def on_resize(self) -> None:
        self._strip_cache.clear()

    def on_mouse_down(self, event: events.Click) -> None:
        if self.zoom_timer:
            self.zoom_timer.stop()
        self.zoom_position = event.offset
        self.zoom_scale = 0.95 if event.ctrl else 1.05
        self.zoom_timer = self.set_interval(1 / 20, self.zoom)
        self.capture_mouse()

    def on_mouse_up(self, event: events.Click) -> None:
        self.release_mouse()
        if self.zoom_timer:
            self.zoom_timer.stop()

    def on_mouse_move(self, event: events.MouseMove) -> None:
        self.zoom_position = event.offset

    def zoom(self) -> None:
        zoom_x, zoom_y = self.zoom_position
        width, height = self.content_size
        x_min, x_max, y_min, y_max = self.set_region

        set_width = x_max - x_min
        set_height = y_max - y_min

        x = x_min + (zoom_x / width) * set_width
        y = y_min + (zoom_y / height) * set_height

        self.set_region = self.set_region.zoom(x, y, self.zoom_scale)

    def notify_style_update(self) -> None:
        self._strip_cache.clear()
        return super().notify_style_update()

    def watch_set_region(self) -> None:
        self._strip_cache.clear()
        self.refresh()

    def render_line(self, y: int) -> Strip:
        if (cached_line := self._strip_cache.get(y)) is not None:
            return cached_line

        width, height = self.content_size
        x_min, x_max, y_min, y_max = self.set_region
        mandelbrot_width = x_max - x_min
        mandelbrot_height = y_max - y_min

        mandelbrot = self.mandelbrot

        max_iterations = self.max_iterations
        set_width = width * 2
        set_height = height * 4
        BRAILLE_MAP = self.BRAILLE_CHARACTERS
        PATCH_COORDS = self.PATCH_COORDS
        max_color = len(COLORS) - 1

        row = y * 4

        colors: list[tuple[int, int, int]] = []

        segments: list[Segment] = []
        base_style = self.rich_style

        for column in range(0, width * 2, 2):
            braille_key = 0
            for bit, dot_x, dot_y in PATCH_COORDS:
                patch_x: int = column + dot_x
                patch_y: int = row + dot_y
                c_real: float = x_min + mandelbrot_width * patch_x / set_width
                c_imag: float = y_min + mandelbrot_height * patch_y / set_height
                if (
                    iterations := mandelbrot(c_real, c_imag, max_iterations)
                ) < max_iterations:
                    braille_key |= bit
                    colors.append(
                        COLORS[round((iterations / max_iterations) * max_color)]
                    )

            if colors:
                patch_red = 0
                patch_green = 0
                patch_blue = 0
                for red, green, blue in colors:
                    patch_red += red
                    patch_green += green
                    patch_blue += blue

                color_count = len(colors)
                patch_color = RichColor.from_rgb(
                    patch_red // color_count,
                    patch_green // color_count,
                    patch_blue // color_count,
                )
                segments.append(
                    Segment(
                        BRAILLE_MAP[braille_key],
                        base_style + RichStyle.from_color(patch_color),
                    )
                )
                colors.clear()
            else:
                segments.append(Segment(" ", base_style))

        strip = Strip(segments, cell_length=width)
        strip.simplify()
        self._strip_cache[y] = strip
        return strip


if __name__ == "__main__":

    class MApp(App):
        CSS = """
        Screen {
            align: center middle;
            background: $panel;
            Mandelbrot {
                background: black 20%;                
                width: 40;
                height: 16;
            }
        }

        """

        def compose(self) -> ComposeResult:
            yield Mandelbrot()

    app = MApp()
    app.run()
